#include <stdio.h>   /* Standard input/output definitions */
#include <stdlib.h>
#include <string.h>  /* String function definitions */
#include <stdbool.h> /* Typedef for boolean */
#include <unistd.h>  /* UNIX standard function definitions */
#include <fcntl.h>   /* File control definitions */
#include <errno.h>   /* Error number definitions */
#include <termios.h> /* POSIX terminal control definitions */
#include <sys/ioctl.h>

#include "csp.h"

int fd = -1; /* File descriptor for the port */

static boolean CspBarcodeRedundancy;		// Device barcode redundancy setting
static boolean CspContinuousScanning;		// Device continuous scan setting

static char CspVolume;					// Device volume setting
static char CspSysstat;					// Battery status (0 = OK, other = low)
static char CspProtoType;				// Protocol Type
static char CspTlBits[ 8 ];				// Device TL Bits
static char CspUploadCount[ 4 ];		// Device upload count
static char CspSignature[ 8 ];			// Device signature
static char CspUserId[ 8 ];				// Device User ID
static char CspSwVersion[ 8 ];			// Device software version
static char CspDeviceId[ 8 ];			// Device ID
static char CspBarData[ MAXSIZE ];		// uploaded barcode storage
static int  CspNBarcodes = 0;			// Number of barcodes
static char ByteBuffer[ MAXSIZE ];		// temp storage of CSP device response

/*--------------------------------------------
 * Functions for communication with CyberPen
 *--------------------------------------------
 */

// Computes the Longitudinal Redundancy Check on a string of specified length
static char cspLrcCheck( char LrcBytes[], int nMaxLength ) {
	int i;
	unsigned char chLrc = 0;        // initialize LRC value

	// compute the LRC for the entire string...
	for( i = 0; i < nMaxLength; i++ )
		chLrc ^= LrcBytes[ i ];

	return( chLrc );
}

static char cspGetc( void ) {
	char chRead;

	// make sure that a port has been selected...   
	if( fd > -1 ) {
		// a port is selected, can we retrieve any data from it?
		// try to read a character...
		usleep( 50 );
		if( read( fd, &chRead, 1 ) > 0 ) {
			return( chRead );
		} else {
			// got no character so timeout must have been reached
			return( READ_TIMEOUT );
		}
		// let other tasks run while we wait for a character...
	}
   // if we make it this far, no character was read. So...  
   return( COMMUNICATIONS_ERROR );
}

static char cspSendCommand( char *Command, int nMaxLength ) {
	int i, sendcnt;
	char buffer[ 255 ];

	// make sure we have a valid port selected...
	if( fd == -1 ) {
		return( COMMUNICATIONS_ERROR );
	}

	// delay before sending message
	usleep( 120 );

	// clear the receive queue...
	while( read( fd, &buffer, sizeof(buffer) ) > 0 );

	sendcnt = 0;
	do {
		sendcnt++;
		// send the command string...
		for( i = 0; i < nMaxLength; i++ ) {
			write( fd, &Command[ i ], 1 );
		}

		// wait for the reply to be processed, or MAXTIME...
		ByteBuffer[ 0 ] = cspGetc();

	} while( ( ByteBuffer[ 0 ] < 0 ) && ( sendcnt < MAXTIME ) );

	// if response not available, reply with error
	if ( ByteBuffer[ 0 ] < 0 ) {
		return( ByteBuffer[ 0 ] );
	}

	if ( ByteBuffer[ 0 ] != NO_ERROR_ENCOUNTERED ) {
		return ( ByteBuffer[ 0 ] );
	} else {
		return ( STATUS_OK );
	}
}

void cspErrMsg( char errcode ) {
	switch( errcode ) {
		case COMMUNICATIONS_ERROR:
			fprintf( stderr, "Port wasn't opened properly." );
			break;
		case LRC_FAILED:
			fprintf( stderr, "LRCheck failed." );
			break;
		case READ_TIMEOUT:
			fprintf( stderr, "Got timeout while trying to read from pen." );
			break;
	}
	printf( "\n\n" );
}

char cspWakeup( void ) {
	char Commands[] = { INTERROGATE, STX, 0x00, 0x03 };
	char ret;
	int i;

	if( ( ret = cspSendCommand( Commands, sizeof( Commands ) ) ) == STATUS_OK ) {
		// device answered with NO_ERROR_ENCOUNTERED so, we read STX next
		ByteBuffer[ 1 ] = cspGetc();

		// Protocol-Type
		ByteBuffer[ 2 ] = cspGetc();
		CspProtoType = ByteBuffer[ 2 ];

		// 00-Systemstatus ok   (other-Low Battery Indication)
		ByteBuffer[ 3 ] = cspGetc();
		CspSysstat = ByteBuffer[ 3 ];

		// 8 Bytes User ID
		for( i = 4; i < 4 + 8; i++ ) {
			ByteBuffer[ i ] = cspGetc();
			CspUserId[ i - 4 ] = ByteBuffer[ i ];
		}

		// 8 Bytes Device Firmware Version
		for( i = 12; i < 12 + 8; i++ ) {
			ByteBuffer[ i ] = cspGetc();
			CspSwVersion[ i - 12 ] = ByteBuffer[ i ];
		}

		// 00-Null (message terminator)
		if( ( ret = cspGetc() ) == 0 ) {
			ByteBuffer[ 20 ] = 0;
			// Compare LRC-Checksum
			if( cspLrcCheck( ByteBuffer, 21 ) == cspGetc() ) {
				return( STATUS_OK );
			} else {
				return( LRC_FAILED );
			}
		} else {
			return( ret );
		}
	} else {
		return( ret );
	}
}

char cspGetCodes( void ) {
	char Commands[] = { UPLOAD, STX, 0x00, 0x05 };
	char ret;
	int i, j, k;
	CspNBarcodes = k = 0;

	if( ( ret = cspSendCommand( Commands, sizeof( Commands ) ) ) == STATUS_OK ) {
		// device answered with NO_ERROR_ENCOUNTERED so, we read STX next
		ByteBuffer[ 1 ] = cspGetc();

		// 8 Bytes Device ID
		for( i = 2; i < 2 + 8; i++ ) {
			ByteBuffer[ i ] = cspGetc();
			CspDeviceId[ i - 2 ] = ByteBuffer[ i ];
		}

		// Upload Counter (shows 00 01 on a new device)
		for( i = 10; i < 10 + 4; i++ ) {
			ByteBuffer[ i ] = cspGetc();
			CspUploadCount[ i - 10 ] = ByteBuffer[ i ];
		}

		// Length of barcode-String
		char bclength;
		while( ( bclength = cspGetc() ) > 0x00 ) {	// Read Codes until 0x00 message-terminator
			ByteBuffer[ i++ ] = bclength;
			CspNBarcodes++;
			for( j = 0; j < bclength; j++ ) {		// Read bclength Bytes representing one barcode
				// put all Bytes into ByteBuffer and CspBarData
				ByteBuffer[ i++ ] = CspBarData[ k++ ] = cspGetc();
			}
			CspBarData[ k++ ] = 0x00; // Seperate each String with 0x00
		}
		if( bclength < 0 ) {
			return( bclength );
		}
		ByteBuffer[ i ] = 0x00; // add 0x00 (message-terminator) to ByteBuffer

		if( cspLrcCheck( ByteBuffer, i + 1 ) == cspGetc() ) {
			printf( "%i Barcode(s) scanned.\n", CspNBarcodes );
			return( STATUS_OK );
		} else {
			return( LRC_FAILED );
		}

	} else {
		return( ret );
	}
}

char cspClear( void ) {
	char Commands[] = { STX, STX, 0x00, 0x00 };
	char ret;
	if( ( ret = cspSendCommand( Commands, sizeof( Commands ) ) ) == STATUS_OK ) {
		// device answered with NO_ERROR_ENCOUNTERED so, we read STX next
		ByteBuffer[ 1 ] = cspGetc();
		// 00-Null
		ByteBuffer[ 2 ] = cspGetc();

		// Compare LRC-Checksum
		if( cspLrcCheck( ByteBuffer, 3 ) == cspGetc() ) {
			return( STATUS_OK );
		} else {
			return( LRC_FAILED );
		}
	} else {
		return( ret );
	}
}

char cspPOff( void ) {
	char Commands[] = { POWER_DOWN, STX, 0x00, 0x07 };
	char ret;
	if( ( ret = cspSendCommand( Commands, sizeof( Commands ) ) ) == STATUS_OK ) {
		// device answered with NO_ERROR_ENCOUNTERED so, we read STX next
		ByteBuffer[ 1 ] = cspGetc();
		// 00-Null
		ByteBuffer[ 2 ] = cspGetc();

		// Compare LRC-Checksum
		if( cspLrcCheck( ByteBuffer, 3 ) == cspGetc() ) {
			return( STATUS_OK );
		} else {
			return( LRC_FAILED );
		}
	} else {
		return( ret );
	}
}

/*----------------------------------------------
 * Functions for communication with other libs
 *----------------------------------------------
 */
int cspNBarcodes( void ) {
	return( CspNBarcodes );
}

char cspProtoType( void ) {
	return( CspProtoType );
}

char cspSysstat( void ) {
	return( CspSysstat );
}

char *cspUserId( void ) {
	return( CspUserId );
}

char *cspSwVersion( void ) {
	return( CspSwVersion );
}

char *cspBarData( void ) {
	return( CspBarData );
}

/*-------------------------------------------
 * Functions for setting up the serial port
 *-------------------------------------------
 */
int open_port( char *port ) {
	fd = open( port, O_RDWR | O_NOCTTY | O_NDELAY );
	if( fd == -1 ) {
		/*
		* Could not open the port.
		*/
		fprintf( stderr, "Unable to open Port %s - ", port );
		perror( "open_port" );

	} else {
		fcntl( fd, F_SETFL, 0 );
	}

	return( fd );

}

int init_port( void ) {
	struct termios options;
	int status = ( TIOCM_RTS | TIOCM_DTR );

	/*
	 * Get the current options for the port...
	 */
	tcgetattr( fd, &options );

	/*
	 * Disable hardware- and software-flowcontrol
	 */
	options.c_iflag &= ~( BRKINT | ICRNL | INLCR | PARMRK | INPCK | ISTRIP | IXON );

	/*
	 * Raw Input/Output Mode
	 */
	options.c_oflag = 0;
	options.c_lflag &= ~( ECHO | ECHONL | ICANON | IEXTEN | ISIG );

	/*
	 * Set Paritty Checking to 8o1
	 */
	options.c_cflag |= PARENB;
	options.c_cflag |= PARODD;
	options.c_cflag &= ~CSTOPB;
	options.c_cflag &= ~CSIZE;
	options.c_cflag |= CS8;

	options.c_cc[ VMIN ]  = 0;
	options.c_cc[ VTIME ] = 10;	// 1 sec timeout
	
	/*
	 * Set the baud rates to 2400...
	 */
	cfsetispeed( &options, B2400 );
	cfsetospeed( &options, B2400 );

	/*
	* Set the new options for the port...
	*/
	if( tcsetattr( fd, TCSANOW, &options ) == -1 ) {
		fprintf( stderr, "Could not set terminal attributes for serial port!\n" );
		return( -1 );
    }

	/*
	 * Set DTS and RTS to high, so dockingstation
	 * gets powered up
	 */
	if( ioctl( fd, TIOCMBIS, &status ) == -1 ) {
		fprintf( stderr, "Unable die set DTS and RTS on serial port!\n" );
		perror( "ioctl()" );
		return( -1 );
    }

	return( 0 );
}

int close_port( void ) {
	close( fd );
	return( 0 );
}
