#include <stdio.h>        /* Standard input/output definitions */
#include <stdlib.h>
#include <string.h>       /* String function definitions */
#include <stdbool.h>      /* Typedef for boolean */
#include <unistd.h>       /* UNIX standard function definitions */
#include <fcntl.h>        /* File control definitions */
#include <errno.h>        /* Error number definitions */
#include <termios.h>      /* POSIX terminal control definitions */
#include <sys/ioctl.h>
#include <signal.h>       /* Signal-handling */
#include <iniparser.h>    /* Library for parsing INI-file */

#include "main.h"
#include "output.h"       /* Library for managing the output - file/uinput */
#include "csp.h"          /* Library for accessing the CyberPen Dockingstation */


/**
 * Copyright (c) 2011-2012 by Moritz Meintker - http://thinksilicon.de
 * 
 * Please see LICENCE and README provided in this package
 */
char *progname;

static void doSigTerm( int sig ) {
	close_port();
	close_output();
	signal( sig, SIG_DFL );
	raise( sig );
}

static boolean file_exists( const char *filename ) {
	FILE *file;
	if( ( file = fopen( filename, "r" ) ) >= 0 ) {
		fclose( file );
		return( true );
	}
	return( false );
}

int main( int argc, char *argv[] ) {
	char *ser_port = "/dev/ttyS0";		// default port to open
	dictionary *settings;

	progname = argv[ 0 ];
	
	boolean en_read  = false;			// Read barcodes of device
	boolean en_clear = false;			// Clear Stick (after reading)
	boolean en_poff  = false;			// Power down (after reading)

	/*-------------------------------------
	 * Handling of command-line-arguments
	 *-------------------------------------
	 */
	int c;
	extern char *optarg;
	extern int optind, optopt, opterr;
	while( ( c = getopt( argc, argv, ":i:p:o:u::rcsh" ) ) != -1 ) {
		switch( c ) {
			case 'i':
				/*
				 * Read given INI-file
				 */
				if( file_exists( optarg ) == true ) {
					settings = iniparser_load( optarg );
					printf( "Reading configuration from %s\n", optarg );

					// get port
					ser_port = iniparser_getstring( settings, "IO:ser_dev", "" );

					// get file-output
					if( iniparser_getboolean( settings, "IO:write_file", 0 ) == true ) {
						char *file_name = iniparser_getstring( settings, "IO:file_name", "" );
						if( open_output_file( file_name ) == -1 ) {
							fprintf( stderr, "Exitting...\n\n" );
							return( 1 );
						} else {
							printf( "Writing output to %s.\n", file_name );
						}
					}

					// get keyboard-output
					if( iniparser_getboolean( settings, "uinput:key_out", 0 ) == true ) {
						printf( "Codes to Keyboard selected.\n" );
						if( open_output_uinput( iniparser_getstring( settings, "uinput:key_device", NULL ) ) == -1 ) {
							fprintf( stderr, "Exitting...\n\n" );
							return( 1 );
						} else {
							// Set the keycodes to be pressed between each code and at the end
							set_delimiter_key( iniparser_getint( settings, "uinput:key_delimiter", 0 ) );
							set_endkey( iniparser_getint( settings, "uinput:key_endkey", 0 ) );
						}
					}

					// get read command
					if( iniparser_getboolean( settings, "Commands:read_codes", 0 ) == true ) {
						en_read = true;
					}

					// get clear command
					if( iniparser_getboolean( settings, "Commands:send_clear", 0 ) == true ) {
						en_clear = true;
					}

					// get poff command
					if( iniparser_getboolean( settings, "Commands:send_poff", 0 ) == true ) {
						en_poff = true;
					}

					printf( "\n" );
				} else {
					fprintf( stderr, "Given INI-file could not be opened.\nExitting...\n\n" );
					return( 1 );
				}
				
				break;
			case 'p':
				/*
				 * Set port for communication with dockingstation
				 */
				ser_port = optarg;
				printf( "Using Serial Port %s\n", ser_port );
				break;
			case 'r':
				/*
				 * Enable read command
				 */
				en_read = true;
				break;
			case 'c':
				/*
				 * Enable clear command
				 */
				en_clear = true;
				break;
			case 's':
				/*
				 * Enable shutdown (power-off) command
				 */
				en_poff = true;
				break;
			case 'o':
				/*
				 * Set output-file - standard to stdout
				 */
				printf( "Writing output to %s.\n", optarg );
				if( open_output_file( optarg ) == -1 ) {
					fprintf( stderr, "Exitting...\n\n" );
					return( 1 );
				}
				break;
			case 'u':
				/*
				 * Set output option for uinput keyboard
				 */
				printf( "Codes to Keyboard selected.\n" );
				if( optarg == NULL ) {
					if( open_output_uinput( NULL ) == -1 ) {
						fprintf( stderr, "Exitting...\n\n" );
						return( 1 );
					}
				} else {
					if( open_output_uinput( optarg ) == -1 ) {
						fprintf( stderr, "Exitting...\n\n" );
						return( 1 );
					}
				}
				break;
			case 'h':
				printf( "Usage: %s [options]\n\
	-i INI-file	Filename of INI-file to read with configuration\n\
	           	Values in this file can be overwritten by command line\n\
	-p Port		Set the serial port to open (default: /dev/ttyS0)\n\
	-o Filename	Set Filename to write output to (default: stdout)\n\
	-u [port]	Keyboard-output via 'uinput'. In order to use this\n\
	           	output, you have to load the uinput-module and make\n\
	           	sure, the device is accessible for this prorgam.\n\
	           	It will look in /dev, /dev/input and /dev/misc for\n\
	           	the uinput device, if it's at an other place on your\n\
	           	machine you can provide a different portname.\n\
	           	After communication with the pen, this prognam will\n\
	           	pause for another 3 seconds to let you change\n\
	           	input-focus to the desired program.\n\
	-r 		Read barcodes (default: reads only UserID, SysStat, etc.)\n\
	-c 		Send clear command after reading\n\
	-s 		Send power-off command after reading\n\
	-h 		prints this help message and exits\n\n", progname );

				return( 0 );
				break;
			case ':':
				fprintf( stderr, "-%c missing argument\n", optopt );
				return( 1 );
				break;
			case '?':
				fprintf( stderr, "unknown argument %c\n", optopt );
				return( 1 );
				break;
		}
	}



	/*----------------------------
	 * Main part starts here
	 *----------------------------
	 */

	if( open_port( ser_port ) ) {
		signal( SIGTERM, doSigTerm );		// Close ports and devices on SIGTERM
		char ret;							// return_values for functions


		init_port();

		printf( "Sending Wakeup-Command.\n" );
		if( ( ret = cspWakeup() ) == STATUS_OK ) {
			printf( "success.\n\n" );
			write_output( STATUS_MSG );
		} else {
			cspErrMsg( ret );

			// Can't initialize communication, this is fatal, so wie quit
			close_port();
			close_output();
			return( 1 );
		}

		if( en_read == true ) {
			printf( "Reading barcodes...   " );
			if( ( ret = cspGetCodes() ) == STATUS_OK ) {
				printf( "success.\n\n" );
			} else {
				printf( "\n" );
				en_read = false;	// set to false so no output will be generated
				cspErrMsg( ret );
			}
		}
		
		if( en_clear == true ) {
			printf( "Sending Clear-Command.\n" );
			if( ( ret = cspClear() ) == STATUS_OK ) {
				printf( "success.\n\n" );
			} else {
				cspErrMsg( ret );
			}
		}
		
		if( en_poff == true ) {
			printf( "Sending Power-Off-Command.\n" );
			if( ( ret = cspPOff() ) == STATUS_OK ) {
				printf( "success.\n\n" );
			} else {
				cspErrMsg( ret );
			}
		}

		close_port();

		if( en_read == true ) {
			write_output( BARCODES );
		}

		close_output();
		return( 0 );
	} else {
		fprintf( stderr, "Exitting...\n\n" );
		return( 1 );
	}
}
