#include <stdio.h>        /* Standard input/output definitions */
#include <stdlib.h>
#include <string.h>       /* String function definitions */
#include <stdbool.h>      /* Typedef for boolean */
#include <unistd.h>       /* UNIX standard function definitions */
#include <fcntl.h>        /* File control definitions */
#include <errno.h>        /* Error number definitions */
#include <linux/input.h>  /* Input subsystem interface */
#include <linux/uinput.h>
#include <time.h>         /* Time-Functions and structs */
#include <sys/times.h>

#include "output.h"
#include "csp.h"          /* Library for accessing the CyberPen Dockingstation */
#include "keys2codes.h"   /* KeysToCodes */


static int   uinp_fd = -1;
static FILE *file_fd;

static boolean file_output   = false;
static boolean uinput_output = false;

static int endkey = 0;
static int delimiter_key = 15;		// Keycode for KEY_TAB

static int Key2Keycode( unsigned char key ) {
	if( key < 128 && KeysToCodes[ key ] != 0 ) {
		return( KeysToCodes[ key ] );
	} else {
		return( -1 );
	}
}

static int Key2Press( unsigned char key ) {
	if( key < 128 && KeysToPress[ key ] != 0 ) {
		return( KeysToPress[ key ] );
	} else {
		return( -1 );
	}
}

static void send_key_event( unsigned int keycode, int keyvalue ) {
	// Check if uinput has been enabled
	if( uinput_output == true && uinp_fd > 0 ) {
		struct input_event event;
		gettimeofday( &event.time, NULL );

		event.type = EV_KEY;
		event.code = keycode;
		event.value = keyvalue;

		if( write( uinp_fd, &event, sizeof( event ) ) < 0 ) {
			printf( "simulate key error\n" );
		}

		event.type = EV_SYN;
		event.code = SYN_REPORT;
		event.value = 0;

		if( write( uinp_fd, &event, sizeof( event ) ) < 0 ) {
			printf( "simulate key error\n" );
		}
	}
}

static char BarCodes_output( FILE *output ) {
	// Check if any codes are available
	if( cspNBarcodes() > 0 ) {
		int i, j;
		char *BarCodes;
		BarCodes = cspBarData();
		j = 0;
		fprintf( output, "Nr.\tType\tCode\n" );

		for( i = 0; i < cspNBarcodes(); i++ ) {
			// Print number and code-type as hex
			fprintf( output, "%i\t0x%X\t", i + 1, BarCodes[ j++ ] );
			while( BarCodes[ j ] > 0x00 && j < MAXSIZE ) {
				fprintf( output, "%c", BarCodes[ j++ ] );
			}
			fprintf( output, "\n" );
			j++;	// next step, so type will be read...
		}
		fprintf( output, "\n" );
	} else {
		return( NOBC_INIT );
	}

	return( STATUS_OK );
}

static char BarCodes_uinput() {
	// Check if any codes are available
	if( cspNBarcodes() > 0 ) {
		int i, j;
		char *BarCodes;
		BarCodes = cspBarData();
		j = 1;

		for( i = 0; i < cspNBarcodes(); i++ ) {
			while( BarCodes[ j ] > 0x00 && j < MAXSIZE ) {

				// Translate Char to Keycode
				if( Key2Keycode( BarCodes[ j ] ) > 0 ) {

					if( Key2Press( BarCodes[ j ] ) > 0 ) {
						send_key_event( Key2Press( BarCodes[ j ] ), 1 );
//						usleep( 8000 );
					}

					send_key_event( Key2Keycode( BarCodes[ j ] ), 1 );	// Button pressed
					send_key_event( Key2Keycode( BarCodes[ j ] ), 0 );	// Button released

					if( Key2Press( BarCodes[ j ] ) > 0 ) {
						send_key_event( Key2Press( BarCodes[ j ] ), 0 );
					}
				} else {
					fprintf( stderr, "Unknown character %X0x\n", BarCodes[ j ] );
				}
				j++;	// read next character

				// make a little sleep here so OOo won't miss keys
				usleep( 8000 );
			}

			// Don't press after the last code
			if( i < cspNBarcodes() - 1 && delimiter_key != 0 ) {
				send_key_event( delimiter_key, 1 );		// Press Delimiter-Key after each code (default KEY_TAB)
				send_key_event( delimiter_key, 0 );
			}

			usleep( 8000 );

			j = j + 2; /* get over next to values, which are the
			            * 0x00 separator and the type. Those we
			            * don't need for uinput output.
			            */
		}

		if( endkey != 0 ) {
			send_key_event( endkey, 1 );	// Press end-Key
			send_key_event( endkey, 0 );
		}
	} else {
		return( NOBC_INIT );
	}

	return( STATUS_OK );
}

void set_delimiter_key( int keycode ) {
	if( keycode >= 0 && keycode < 0x300 ) {
		delimiter_key = keycode;
	}
}

void set_endkey( int keycode ) {
	if( keycode >= 0 && keycode < 0x300 ) {
		endkey = keycode;
	}
}

char open_output_file( char *filename ) {
	// open file for writing
	file_fd = fopen( filename, "w" );
	if( file_fd == NULL ) {
		/*
		* Could not open filename.
		*/
		fprintf( stderr, "Unable to open File %s for writing - ", filename );
		perror( "open_output" );
		return( -1 );
	} else {
		file_output = true;
		return( 0 );
	}
}

char open_output_uinput( char *uinp_manualdev ) {
	// logic for opening uinput device
	char *uinp_device[] = { "/dev/uinput", "/dev/input/uinput", "/dev/misc/uinput" };
	int i = 0;

	if( uinp_manualdev == NULL ) {
		for( i = 0; i < 3; i++ ) {
			// Try to open the uinput device
			if( ( uinp_fd = open( uinp_device[ i ], O_WRONLY | O_NDELAY ) ) <= 0 ) {
				continue;
			} else break;
		}
	} else {
		uinp_fd = open( uinp_manualdev, O_WRONLY | O_NDELAY );
	}

	if( uinp_fd == -1 ) {
		/*
		* Could not open filename.
		*/
		perror( "Unable to open uinput device - open_output" );
		return( -1 );
	}

	printf( "Using " );
	if( uinp_manualdev == NULL ) {
		printf( "%s\n", uinp_device[ i ] );
	} else {
		printf( "%s\n", uinp_manualdev );
	}

	struct uinput_user_dev uinp;
	memset( &uinp, 0, sizeof( uinp ) ); // Intialize the uInput device to NULL
	strncpy( uinp.name, "Virtual Keyboard Device", strlen( "Virtual Keyboard Device" ) );
	uinp.id.version = 4;
	uinp.id.bustype = BUS_USB;
	uinp.id.product = 1;
	uinp.id.vendor = 1;

	/*
		* Setup the uinput device
		*/
	ioctl( uinp_fd, UI_SET_EVBIT, EV_KEY );
	ioctl( uinp_fd, UI_SET_EVBIT, EV_REL );
	for( i = 0; i < 256; i++ ) {
		ioctl( uinp_fd, UI_SET_KEYBIT, i );
	}

	/*
		* Create input device into input sub-system
		*/
	write( uinp_fd, &uinp, sizeof( uinp ) );
	if( ioctl( uinp_fd, UI_DEV_CREATE ) ) {
		printf( "Unable to create UINPUT device." );
		return( -1 );
	} else {
		uinput_output = true;
		return( 0 );
	}
}

char write_output( int output ) {

	/*
	 * Print to stdout
	 */
	switch( output ) {
		case STATUS_MSG:
			printf( "Wakeup returned:\n\nPType:  %X\tSystemstatus: %X\nUserID: %s\tFirmware:     %s\n\n",\
			         cspProtoType(), cspSysstat(), cspUserId(), cspSwVersion() );

			break;
		case BARCODES:
			BarCodes_output( stdout );
			break;
	}
	

	if( file_output == true ) {
		/*
		 * Write to file
		 */
		switch( output ) {
			case STATUS_MSG:
				fprintf( file_fd, "Wakeup returned:\n\nPType:   %X\tSystemstatus: %X\nUserID: %s\tFirmware:     %s\n\n",\
				        cspProtoType(), cspSysstat(), cspUserId(), cspSwVersion() );
				break;
			case BARCODES:
				BarCodes_output( file_fd );
				break;
		}
	}

	if( uinput_output == true ) {
		/*
		 * Send to uinput device
		 */
		switch( output ) {
			case BARCODES:
				sleep( 3 );
				BarCodes_uinput();
				break;
		}
	}

	return( 0 );
}

char close_output() {
	if( file_output == true ) {
		return(  fclose( file_fd ) );
	}
	if( uinput_output == true ) {
		// logic for closing uinput device
		if( ioctl( uinp_fd, UI_DEV_DESTROY ) == -1 ) {
			perror( "ioctl UI_DEV_DESTROY" );
			return( -1 );
		}
		return( close( uinp_fd ) );
	}
	return( 0 );
}
